/*
 * main.c
 *
 * Copyright (C) 1993, 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <sys/param.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "common.h"
#include "StrReq.h"
#include "Bookmark.h"
#include "WWW.h"
#include "resources.h"

struct data
{
  char *data;
  int len;
};

static void sigigh_handler _ArgProto(());
static Widget CreateStrReq _ArgProto((char *, char *, int,
				      XtCallbackProc, XtCallbackProc,
				      XtPointer));
static void xtWarningHandler _ArgProto((String));
static void xtErrorHandler _ArgProto((String));
static int xErrorHandler _ArgProto((Display *, XErrorEvent *));

static void OpenURL _ArgProto(());

AppResources appr;

#define BUTTON_LIST "quit, open, home, back, source, reload, file, help, bookmark, search, cancel"

#define offset(field) XtOffset(AppResources *, field)
static XtResource       resource_list[] =
{
  { "homeURL", "URL", XtRString, sizeof(char *),
	offset(homeURL), XtRString, (XtPointer)"file://localhost/" },
  { "helpURL", "URL", XtRString, sizeof(char *),
	offset(helpURL), XtRString, (XtPointer)"file://localhost/" },
  { "showURL", XtCBoolean, XtRBoolean, sizeof(Boolean),
	offset(showURL), XtRImmediate, (XtPointer)True },
  { "button1Box", "BoxList", XtRString, sizeof(char *),
	offset(button1Box), XtRString, (XtPointer)BUTTON_LIST },
  { "button2Box", "BoxList", XtRString, sizeof(char *),
	offset(button2Box), XtRString, (XtPointer)NULL },
  { "printerName", "PrinterName", XtRString, sizeof(char *),
	offset(printerName), XtRString, (XtPointer)"lp" },
  { "bookmarkFile", XtCFile, XtRString, sizeof(char *),
	offset(bookmarkFile), XtRString, (XtPointer)"~/.chimera_bookmark" },
  { XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
	offset(messageFont), XtRString, "variable" },
  { "sourceViewer", "SourceViewer", XtRString, sizeof(char *),
	offset(sourceViewer), XtRString, (XtPointer)"axe -noserver" },
};

extern char *fallback_resources[];

static XtActionsRec actionsList[] =
{
  { "openurl",  (XtActionProc) OpenURL },
  { "quit",  (XtActionProc) Quit },
};

/*
 * main
 *
 * this is where the whole thing kicks off.
 */
void
main(argc, argv)
int argc;
char **argv;
{
  char *first;
  Atom delete;
  XGCValues xgcv;

#ifdef SOCKS
  SOCKSinit(argv[0]);
#endif

  signal(SIGINT, sigigh_handler);
  signal(SIGQUIT, sigigh_handler);
  signal(SIGHUP, sigigh_handler);
  signal(SIGTERM, sigigh_handler);
  signal(SIGPIPE, SIG_IGN);

  StartReaper();

  appr.toplevel = XtAppInitialize(&(appr.appcon), "Chimera", NULL, 0,
				  &argc, argv, fallback_resources, NULL, 0);
  if (appr.toplevel == 0) exit(1);

  XtGetApplicationResources(appr.toplevel, &appr, resource_list,
			    XtNumber(resource_list), NULL, 0);

  /*
   * this is hiding in widget.c.
   */
  CreateWidgets(&appr);

  XtRealizeWidget(appr.toplevel);

  XtAppAddActions(appr.appcon, actionsList, XtNumber(actionsList));
  if (appr.urldisplay != 0)
  {
    XtOverrideTranslations(appr.urldisplay,
			   XtParseTranslationTable("<Key>Return: openurl()"));
  }

  xgcv.font = appr.messageFont->fid;
  xgcv.foreground = BlackPixel(XtDisplay(appr.toplevel),
                               DefaultScreen(XtDisplay(appr.toplevel)));
  appr.gc = XCreateGC(XtDisplay(appr.toplevel), XtWindow(appr.message),
                      GCFont | GCForeground, &xgcv);

  XtAppSetErrorHandler(appr.appcon, xtErrorHandler);
/*  XtAppSetWarningHandler(appr.appcon, xtWarningHandler); */
  XSetErrorHandler(xErrorHandler);

  delete = XInternAtom(XtDisplay(appr.toplevel), "WM_DELETE_WINDOW", False);
  XtOverrideTranslations (appr.toplevel, 
			  XtParseTranslationTable
			  ("<Message>WM_PROTOCOLS: quit()"));
  XSetWMProtocols (XtDisplay(appr.toplevel), XtWindow(appr.toplevel),
		   &delete, 1);

  /*
   * Initialize AppResources
   */
  appr.group = NULL;
  appr.bookpop = NULL;
  appr.openpop = NULL;

  /*
   * Grab the first document.  First, build a fake context URL in case the
   * user provides a relative URL.  Create a new URL based on the user
   * supplied URL and then fake context and try to load it.  If that
   * fails then supply a document that is just an error message.
   */
  if (argc > 1) first = argv[1];
  else if ((first = getenv("WWW_HOME")) != NULL) ;
  else first = appr.homeURL;

  XtVaSetValues(appr.www, XtNurl, first, NULL);

  XtAppMainLoop(appr.appcon);
}

/*
 * ClearMessage
 */
static void
ClearMessage()
{
  XClearWindow(XtDisplay(appr.message), XtWindow(appr.message));
  XFlush(XtDisplay(appr.message));

  return;
}

/*
 * SetMessage
 */
static void
SetMessage(str)
char *str;
{
  Display *dpy;

  ClearMessage();

  dpy = XtDisplay(appr.message);
  XDrawString(dpy, XtWindow(appr.message), appr.gc,
	      10, appr.messageFont->ascent + 2, str, strlen(str));
  XFlush(dpy);

  return;
}

/*
 * DisplayCallback
 */
void
DisplayCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  if (appr.showURL)
  {
    XtVaSetValues(appr.urldisplay, XtNstring, (char *)cbdata, NULL);
    XFlush(XtDisplay(appr.toplevel));
  }

  ClearMessage();

  return;
}

/*
 * DownloadCallback
 */
void
DownloadCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  static char *down = "Downloading ";
  char *message;

  message = (char *)alloc_mem(strlen(down) + strlen((char *)cbdata) + 1);
  strcpy(message, down);
  strcat(message, (char *)cbdata);
  SetMessage(message);
  free_mem(message);

  return;
}

/*
 * AnchorCallback
 */
void
AnchorCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  if (cbdata == NULL) ClearMessage();
  else SetMessage((char *)cbdata);

  return;
}

/*
 * CreateStrReq
 */
static Widget
CreateStrReq(s, defstr, hidden, ocallback, dcallback, cldata)
char *s;
char *defstr;
int hidden;
XtCallbackProc ocallback;
XtCallbackProc dcallback;
XtPointer cldata;
{
  Widget nw;
  Arg args[10];
  int argcnt;
  Window rw, cw;
  int rx, ry, wx, wy;
  unsigned int mask;

  XQueryPointer(XtDisplay(appr.toplevel), XtWindow(appr.toplevel),
		&rw, &cw,
		&rx, &ry,
		&wx, &wy,
		&mask);

  argcnt = 0;
  XtSetArg(args[argcnt], XtNdefaultString, defstr); argcnt++;
  XtSetArg(args[argcnt], XtNhiddenText, hidden); argcnt++;
  XtSetArg(args[argcnt], XtNx, rx - 2); argcnt++;
  XtSetArg(args[argcnt], XtNy, ry - 2); argcnt++;
  nw = XtCreatePopupShell(s,
			  strReqWidgetClass, appr.toplevel,
			  args, argcnt);
  XtAddCallback(nw, XtNokCallback, ocallback, cldata);
  if (dcallback != NULL)
  {
    XtAddCallback(nw, XtNdismissCallback, dcallback, cldata);
  }

  XtRealizeWidget(nw);

  return(nw);
}

/*
 * SOCallback
 */
static void
SOCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  struct data *d = (struct data *)cldata;
  FILE *fp;

  if (cbdata != NULL)
  {
    if ((fp = fopen(cbdata, "w")) != NULL)
    {
      fwrite(d->data, d->len, 1, fp);
      fclose(fp);
    }
  }

  free_mem((char *)d->data);
  free_mem((char *)d);

  XtPopdown(w); 
  XtDestroyWidget(w);

  return;
}

/*
 * SDCallback
 */
static void
SDCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  struct data *d = (struct data *)cldata;

  free_mem((char *)d->data);
  free_mem((char *)d);

  XtPopdown(w); 
  XtDestroyWidget(w);

  return;
}

/*
 * SaveCallback
 *
 * This really, really sucks.  What if someone wants to save a 10MB file?
 * Yipe!
 */
void
SaveCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWCallbackInfo *wwwci = (WWWCallbackInfo *)cbdata;
  struct data *d;

  d = (struct data *)alloc_mem(sizeof(struct data));
  d->data = (char *)alloc_mem(wwwci->len);
  memcpy(d->data, wwwci->data, wwwci->len);
  d->len = wwwci->len;

  CreateStrReq("filename", NULL, False, SOCallback, SDCallback, d);

  return;
}

/*
 * OOCallback
 */
static void
OOCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  if (cbdata != NULL)
  {
    XtVaSetValues(appr.www, XtNurl, (char *)cbdata, NULL);
    XtPopdown(w); 
  }

  return;
}

/*
 * OpenURL
 */
static void
OpenURL()
{
  char *s;

  XtVaGetValues(appr.urldisplay, XtNstring, &s, NULL);
  XtVaSetValues(appr.www, XtNurl, s, NULL);

  return;
}

/*
 * Open
 */
void
Open(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  if (appr.openpop == NULL)
  {
    appr.openpop = CreateStrReq("url", NULL, False, OOCallback, NULL, 0);
  }

  XtPopup(appr.openpop, XtGrabNone);

  return;
}

/*
 * Home
 */
void
Home(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWHome(appr.www);
  return;
}

/*
 * Back
 */
void
Back(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWBack(appr.www);
  return;
}

/*
 * Reload
 */
void
Reload(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWReload(appr.www);
  return;
}

/*
 * Cancel
 */
void
Cancel(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWCancel(appr.www);
  return;
}

/*
 * Help
 *
 * Shows the help file.
 */
void
Help(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  XtVaSetValues(appr.www, XtNurl, appr.helpURL, NULL);
  return;
}

/*
 * Search
 */
void
Search(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWSearch(appr.www, "", True);
  return;
}

/*
 * File
 */
void
File(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  return;
}

/*
 * Quit
 */
void
Quit(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  XtDestroyWidget(appr.www);
  exit(0);
}

/*
 * sigiqh_handler
 */
static void
sigigh_handler()
{
  Quit(appr.quit, &appr, NULL);
  return;
}

/*
 * Source
 *
 * Oh boy does this suck!
 */
void
Source(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  int len;
  char *data;
  FILE *fp;
  char buffer[L_tmpnam + 1];
  char *fname;
  char *c;

  if ((fname = tmpnam(buffer)) == NULL) return;

  XtVaGetValues(appr.www,
		XtNsource, &data,
		XtNsourceLength, &len,
		NULL);

  if ((fp = fopen(fname, "w")) != NULL)
  {
    fwrite(data, len, 1, fp);
    fclose(fp);

    if (fork() == 0)
    {
      c = (char *)alloc_mem(strlen(appr.sourceViewer) + strlen(fname) + 2);
      strcpy(c, appr.sourceViewer);
      strcat(c, " ");
      strcat(c, fname);
      system(c);
      free_mem(c);
      unlink(fname);
      exit(0);
    }
  }

  return;
}

/*
 * BookmarkCallback
 */
static void
BookmarkCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  if (cbdata != NULL)
  {
    XtVaSetValues(appr.www, XtNurl, (char *)cbdata, NULL);
    XtPopdown(w);
  }

  return;
}

/*
 * MyBookmarkProc
 */
static void
MyBookmarkProc(cldata, display, url)
XtPointer cldata;
char **display;
char **url;
{
  char *s1, *s2;

  XtVaGetValues(appr.www,
		XtNtitle, &s1,
		XtNurl, &s2,
		NULL);

  if (s1 == NULL) s1 = "";
  if (s2 == NULL) s2 = "";

  *display = alloc_string(s1);
  *url = alloc_string(s2);

  return;
}

/*
 * BookmarkDestroyCallback
 */
static void
BookmarkDestroyCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  String g;

  XtVaGetValues(w, XtNcurrentGroup, &g, NULL);
  if (g != NULL)
  {
    if (appr.group != NULL) free_mem(appr.group);
    appr.group = alloc_string(g);
  }

  return;
}

/*
 * Bookmark
 */
void
Bookmark(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  Arg args[10];
  int argcnt;
  Window rw, cw;
  int rx, ry, wx, wy;
  unsigned int mask;
  char *filename;

  if (appr.bookpop == 0)
  {
    filename = FixFilename(appr.bookmarkFile);
    
    XQueryPointer(XtDisplay(appr.toplevel), XtWindow(appr.toplevel),
		  &rw, &cw,
		  &rx, &ry,
		  &wx, &wy,
		  &mask);
    
    argcnt = 0;
    XtSetArg(args[argcnt], XtNbookmarkProc, MyBookmarkProc); argcnt++;
    XtSetArg(args[argcnt], XtNbookmarkProcData, 0); argcnt++;
    XtSetArg(args[argcnt], XtNfilename, filename); argcnt++;
    XtSetArg(args[argcnt], XtNx, rx); argcnt++;
    XtSetArg(args[argcnt], XtNy, ry); argcnt++;
    if (appr.group != NULL)
    {
      XtSetArg(args[argcnt], XtNcurrentGroup, appr.group); argcnt++;
    }
    appr.bookpop = XtCreatePopupShell("bookmark",
				      bookmarkWidgetClass, appr.toplevel,
				      args, argcnt);
    XtAddCallback(appr.bookpop, XtNcallback, BookmarkCallback, 0);
    XtAddCallback(appr.bookpop, XtNdestroyCallback, BookmarkDestroyCallback,
		  (XtPointer)&appr);
    
    XtRealizeWidget(appr.bookpop);
  }

  XtPopup(appr.bookpop, XtGrabNone);

  return;
}

/*
 * xErrorHandler
 */
static int
xErrorHandler(dpy, xe)
Display *dpy;
XErrorEvent *xe;
{
  fprintf (stderr, "X error\n");
  fflush(stderr);
  return(0);
}

/*
 * xtErrorHandler
 */
static void
xtErrorHandler(msg)
String msg;
{
  fprintf (stderr, "%s\n", msg);
  fflush(stderr);
  return;
}

/*
 * xtWarningHandler
 */
static void
xtWarningHandler(msg)
String msg;
{
  fprintf (stderr, "%s\n", msg);
  fflush(stderr);
  return;
}
