/*
 * widget.c
 *
 * Copyright (C) 1993, 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * Note that most of this code was actually written by Jim.Rees@umich.edu.
 * The messed up parts were written by john.
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Toggle.h>

#include "ScrollText.h"

#include "common.h"
#include "WWW.h"
#include "resources.h"

#define WT_COMMAND 0
#define WT_TOGGLE 1

static struct ButtonTable
{
  char *name;
  Widget w;
  void (*cb)();
  int type;
} ButtonTable[] =
{
  {"quit",      NULL, Quit, WT_COMMAND },
  {"open",	NULL, Open, WT_COMMAND },
  {"home",	NULL, Home, WT_COMMAND },
  {"back",	NULL, Back, WT_COMMAND },
  {"source",	NULL, Source, WT_COMMAND },
  {"reload",	NULL, Reload, WT_COMMAND },
  {"file",      NULL, File, WT_COMMAND },
  {"help",	NULL, Help, WT_COMMAND },
  {"cancel",    NULL, Cancel, WT_COMMAND },
  {"bookmark",	NULL, Bookmark, WT_COMMAND },
  {"search",    NULL, Search, WT_COMMAND },
  {NULL,        NULL, NULL },
};

static void
AddButtons(r, box, list)
AppResources *r;
Widget box;
char *list;
{
  char name[256];
  struct ButtonTable *btp;

  list = alloc_string(list);

  while (sscanf(list, " %[^,]", name) == 1)
  {
    /*
     * Find the listed button and create its widget
     */
    for (btp = &ButtonTable[0]; btp->name != NULL; btp++)
    {
      if (!strcasecmp(btp->name, name))
      {
	btp->w = XtCreateManagedWidget(btp->name,
				       btp->type == WT_COMMAND ?
				       commandWidgetClass:toggleWidgetClass,
				       box,
				       NULL, 0);
	XtAddCallback(btp->w, XtNcallback, btp->cb, r);
	break;
      }
    }

    /*
     * Fill in the entry in AppResources for the widget.
     */
    if (!strcasecmp(name, "open")) r->open = btp->w;
    else if (!strcasecmp(name, "home")) r->home = btp->w;
    else if (!strcasecmp(name, "back")) r->back = btp->w;
    else if (!strcasecmp(name, "source")) r->source = btp->w;
    else if (!strcasecmp(name, "reload")) r->reload = btp->w;
    else if (!strcasecmp(name, "file")) r->file = btp->w;
    else if (!strcasecmp(name, "help")) r->help = btp->w;
    else if (!strcasecmp(name, "bookmark")) r->bookmark = btp->w;
    else if (!strcasecmp(name, "cancel")) r->cancel = btp->w;
    else if (!strcasecmp(name, "search")) r->search = btp->w;
    else if (!strcasecmp(name, "quit")) r->quit = btp->w;

    /*
     * Skip to the next comma-delimited item in the list
     */
    while (*list && *list != ',')
      list++;
    if (*list == ',')
      list++;
  }

  return;
}

/*
 * CreateWidgets
 *
 * Create the command and box widgets.
 */
void
CreateWidgets(r)
AppResources *r;
{
  Widget paned, box, form;
  Arg args[10];
  int argcnt;

  /*
   * Main window pane
   */
  paned = XtCreateManagedWidget("paned",
                                panedWidgetClass, r->toplevel, 
                                NULL, 0);


  /*
   * Button pane(s)
   */
  if (r->button1Box && *r->button1Box)
  {
    box = XtCreateManagedWidget("box1", boxWidgetClass, paned, NULL, 0);
    AddButtons(r, box, r->button1Box);
  }

  if (r->button2Box && *r->button2Box)
  {
    box = XtCreateManagedWidget("box3", boxWidgetClass, paned, NULL, 0);
    AddButtons(r, box, r->button2Box);
  }

  /*
   * Third pane.  URL display.
   */
  if (r->showURL)
  {
    form = XtCreateManagedWidget("box3",
				formWidgetClass, paned,
				NULL, 0);
    
    XtCreateManagedWidget("urllabel",
			  labelWidgetClass, form,
			  NULL, 0);
    
    r->urldisplay = XtCreateManagedWidget("urldisplay",
					  scrollingTextWidgetClass, form,
					  NULL, 0);
    r->urldisplay = XtNameToWidget(r->urldisplay, "text");
  }
  else
  {
    r->urldisplay = 0;
  }

  /*
   * Fourth Pane.  General message area.
   */
  argcnt = 0;
  XtSetArg(args[argcnt], XtNheight,
	   r->messageFont->ascent + r->messageFont->descent + 4);
  argcnt++;
  XtSetArg(args[argcnt], XtNwidth, 1); argcnt++;
  r->message = XtCreateManagedWidget("message",
				     widgetClass, paned,
				     args, argcnt);

  /*
   * Fifth pane, WWW widget
   */
  r->www = XtCreateManagedWidget("www",
				 wwwWidgetClass, paned,
				 NULL, 0);
  XtAddCallback(r->www, XtNdisplayCallback, DisplayCallback, (XtPointer)0);
  XtAddCallback(r->www, XtNdownloadCallback, DownloadCallback, (XtPointer)0);
  XtAddCallback(r->www, XtNdocumentSaveCallback, SaveCallback, (XtPointer)0);
  XtAddCallback(r->www, XtNanchorCallback, AnchorCallback, (XtPointer)0);

  return;
}


