/*
 * WWW.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for more information.
 */
#include "copyright.h"

#include <stdio.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Scrollbar.h>
#include "DrawingArea.h"

#include "WWWP.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

#define offset(field) XtOffsetOf(WWWRec, www.field)
static XtResource resources[] =
{
  { XtNurl, XtCURL, XtRString, sizeof(char *),
	offset(url), XtRImmediate, (XtPointer)NULL },

  { XtNh1Font, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(h1Font), XtRString,
	"-sony-fixed-medium-r-normal--24-170-100-100-c-120-iso8859-1" },

  { XtNh2Font, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(h2Font), XtRString,
	"-misc-fixed-medium-r-normal--20-200-75-75-c-100-iso8859-1" },

  { XtNh3Font, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(h3Font), XtRString, (XtPointer)"10x20" },

  { XtNh4Font, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(h4Font), XtRString, (XtPointer)"10x20" },

  { XtNh5Font, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(h5Font), XtRString, (XtPointer)"10x20" },

  { XtNh6Font, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(h6Font), XtRString, (XtPointer)"10x20" },

  { XtNnormalFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(normalFont), XtRString, (XtPointer)"variable" },

  { XtNemFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(emFont), XtRString, (XtPointer)"variable" },

  { XtNbFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(bFont), XtRString, (XtPointer)"variable" },

  { XtNiFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(iFont), XtRString,
	"-adobe-helvetica-medium-o-*-*-*-120-*-*-*-*-iso8859-1"},

  { XtNttFont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(ttFont), XtRString, "fixed" },

  { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(foreground), XtRString, (XtPointer)"black" },

  { XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
	offset(background), XtRString, (XtPointer)"moccasin" },

  { XtNanchorColor, XtCColor, XtRPixel, sizeof(Pixel),
	offset(anchorColor), XtRString, (XtPointer)"blue" },

  { XtNmargin, XtCMargin, XtRDimension, sizeof(Dimension),
	offset(margin), XtRImmediate, (XtPointer)20 },

  { XtNlineSpace, XtCMargin, XtRDimension, sizeof(Dimension),
	offset(lineSpace), XtRImmediate, (XtPointer)2 },

  { XtNanchorCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	offset(anchorCallbacks), XtRCallback, (XtPointer)NULL },

  { XtNdisplayCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	offset(displayCallbacks), XtRCallback, (XtPointer)NULL },

  { XtNdownloadCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	offset(downloadCallbacks), XtRCallback, (XtPointer)NULL },

  { XtNdocumentSaveCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	offset(saveCallbacks), XtRCallback, (XtPointer)NULL },

  { XtNmaxDownloads, XtCMaxDownloads, XtRInt, sizeof(int),
	offset(maxDownloads), XtRImmediate, (XtPointer)3 },

  { XtNanchorColor, XtCColor, XtRPixel, sizeof(Pixel),
	offset(anchorColor), XtRString, (XtPointer)"blue" },

  { XtNanchorReport, XtCBoolean, XtRBoolean, sizeof(Boolean),
	offset(anchorReport), XtRString, (XtPointer)"True" },

  { XtNimageQuitInterval, XtCInterval, XtRInt, sizeof(int),
	offset(imageQuitInterval), XtRImmediate, (XtPointer)10000 },

  { XtNioQuantum, XtCIOQuantum, XtRInt, sizeof(int),
	offset(ioQuantum), XtRImmediate, (XtPointer)256 },

  { XtNioTimeOutInterval, XtCInterval, XtRInt, sizeof(int),
	offset(ioTimeOutInterval), XtRImmediate, (XtPointer)10000 },

  { XtNmimeCapFiles, XtCFileList, XtRString, sizeof(String),
	offset(mimeCapFiles), XtRImmediate, (XtPointer)"~/.mailcap" },

  { XtNmimeTypeFiles, XtCFileList, XtRString, sizeof(String),
	offset(mimeTypeFiles), XtRImmediate, (XtPointer)"~/.mime.types" },

  { XtNpath, XtCFileList, XtRString, sizeof(String),
	offset(path), XtRImmediate, (XtPointer)"." },

  { XtNmaxDownloads, XtCMaxDownloads, XtRInt, sizeof(int),
	offset(maxDownloads), XtRImmediate, (XtPointer)4 },

  { XtNbrokenHTML, XtCBrokenHTML, XtRBoolean, sizeof(Boolean),
	offset(brokenHTML), XtRImmediate, (XtPointer)True },
};
#undef offset

static void WWWInitialize(), WWWResize(), WWWDestroy(), WWWRealize();
static Boolean WWWSetValues();
static XtGeometryResult WWWGeometryManager();
static XtGeometryResult WWWPreferredGeometry();

static void WWWGetValuesHook _ArgProto((Widget, ArgList, Cardinal *));

WWWClassRec wwwClassRec =
{
  { /* core_class fields */
    /* superclass         */    (WidgetClass) &constraintClassRec,
    /* class_name         */    "WWW",
    /* widget_size        */    sizeof(WWWRec),
    /* class_initialize   */    NULL,
    /* class_part_init    */    NULL,
    /* class_inited       */    FALSE,
    /* initialize         */    WWWInitialize,
    /* initialize_hook    */    NULL,
    /* realize            */    WWWRealize,
    /* actions            */    NULL,
    /* num_actions        */    0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion    */    FALSE,
    /* compress_exposure  */    FALSE,
    /* compress_enterleave*/    FALSE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    WWWResize,
    /* expose             */    NULL,
    /* set_values         */    WWWSetValues,
    /* set_values_hook    */    NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */    WWWGetValuesHook,
    /* accept_focus       */    NULL,
    /* version            */    XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    NULL,
    /* query_geometry     */	WWWPreferredGeometry,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
  },
  { /* composite_class fields */
    /* geometry_manager   */   WWWGeometryManager,
    /* change_managed     */   NULL,
    /* insert_child       */   XtInheritInsertChild,
    /* delete_child       */   XtInheritDeleteChild,
    /* extension          */   NULL
  },
  { /* constraint_class fields */
    /* subresourses       */   NULL,
    /* subresource_count  */   0,
    /* constraint_size    */   0,
    /* initialize         */   NULL,
    /* destroy            */   NULL,
    /* set_values         */   NULL,
    /* extension          */   NULL
  },
  { /* form_class fields */
    /* layout             */   NULL
  }
};

WidgetClass wwwWidgetClass = (WidgetClass)&wwwClassRec;

/*
 * WWWGetValuesHook
 */
static void
WWWGetValuesHook(w, args, num_args)
Widget w;
ArgList args;
Cardinal * num_args;
{
  WWWWidget rw = (WWWWidget)w;
  int i;

  for (i = 0; i < *num_args; i++)
  {
    if (strcmp(args[i].name, XtNbackground) == 0)
    {
      *((Pixel *)args[i].value) = rw->www.background;
    }
    if (strcmp(args[i].name, XtNforeground) == 0)
    {
      *((Pixel *)args[i].value) = rw->www.foreground;
    }
    else if (strcmp(args[i].name, XtNsource) == 0)
    {
      if (rw->www.dilist != NULL && rw->www.dilist->d != NULL)
      {
	*((char **)args[i].value) = rw->www.dilist->d->data;
      }
      else *((char **)args[i].value) = NULL;
    }
    else if (strcmp(args[i].name, XtNsourceLength) == 0)
    {
      if (rw->www.dilist != NULL && rw->www.dilist->d != NULL)
      {
	*((int *)args[i].value) = rw->www.dilist->d->len;
      }
    }
    else if (strcmp(args[i].name, XtNtitle) == 0)
    {
      *((char **)args[i].value) = rw->www.title;
    }
    else if (strcmp(args[i].name, XtNauthor) == 0)
    {
      *((char **)args[i].value) = rw->www.author;
    }
    else if (strcmp(args[i].name, XtNurl) == 0)
    {
      *((char **)args[i].value) = rw->www.rurl;
    }
  }

  return;
}

/*
 * WWWResize
 */
static void
WWWResize(w)
Widget w;
{
  WWWWidget rw = (WWWWidget)w;

  ReloadDocument(rw);

  return;
}

/*
 * WWWDestroy
 */
static void
WWWDestroy(w)
Widget w;
{
  WWWWidget rw = (WWWWidget)w;

  XtReleaseGC(w, rw->www.gc);
  if (rw->www.vert_bar != NULL) XtDestroyWidget(rw->www.vert_bar);
  if (rw->www.horiz_bar != NULL) XtDestroyWidget(rw->www.horiz_bar);
  if (rw->www.clip != NULL) XtDestroyWidget(rw->www.clip);
  if (rw->www.child != NULL) XtDestroyWidget(rw->www.child);

  return;
}

/*
 * WWWRealize
 */
static void
WWWRealize(w, valueMask, attributes)
Widget w;
XtValueMask *valueMask;
XSetWindowAttributes *attributes;
{
  (*wwwWidgetClass->core_class.superclass->core_class.realize)
      (w, valueMask, attributes);

  return;
}

static XtGeometryResult
WWWPreferredGeometry(w, con, reply)
Widget w;
XtWidgetGeometry *con, *reply;
{
  WWWWidget rw = (WWWWidget)w;

  if (rw->www.rphead == NULL)
  {
    XtResizeWidget(rw->www.clip, con->width > 0 ? con->width:100,
		   con->height > 0 ? con->height:100, 0);
  }

  return(XtGeometryYes);
}

/*
 * WWWGeometryManager
 */
static XtGeometryResult
WWWGeometryManager(w, request, reply)
Widget w;
XtWidgetGeometry *request;
XtWidgetGeometry *reply;
{ 
  return(XtGeometryNo);
}

/*
 * WWWSetValues
 */
static Boolean
WWWSetValues(c, r, n)
Widget c, r, n;
{
  WWWWidget current = (WWWWidget)c;
  WWWWidget new = (WWWWidget)n;

  if (current->www.url != new->www.url)
  {
    URLParts *up;

    up = ParseURL(new->www.url);
    NewDocument(new, up);
    DestroyURLParts(up);
  }

  return(False);
}

/*
 * GetSaveFilename
 *
 * Original code/idea by Roman with additional mangling by john.
 */
/*
static char *
GetSaveFilename(up)
URLParts *up;
{
  char *fn;
  char *basename;
  int length;
  static char filename[MAXPATHLEN + 1];

  if (up == NULL || up->filename == NULL ||
      (length = strlen(up->filename)) == 0)
  {
    return("");
  }

  strcpy(filename, up->filename);
  if (filename[length - 1] == '/') filename[length - 1] = '\0';
  if ((basename = strrchr (filename, '/')) != NULL) fn = basename + 1;
  else fn = filename;

  return(fn);
}
*/

/*
 * Scroll handling code snarfed from Viewport.c from Xaw.
 */
static void
SetBar(w, top, length, total)
Widget w;
Position top;
Dimension length, total;
{
  XawScrollbarSetThumb(w, (float)top/(float)total,
		       (float)length/(float)total);

  return;
}

static void
RedrawThumbs(rw)
WWWWidget rw;
{
  register Widget child = rw->www.child;
  register Widget clip = rw->www.clip;
  
  if (rw->www.horiz_bar != (Widget)NULL)
  {
    SetBar(rw->www.horiz_bar, -(child->core.x),
	   clip->core.width, child->core.width );
  }
  
  if (rw->www.vert_bar != (Widget)NULL)
  {
    SetBar(rw->www.vert_bar, -(child->core.y),
	   clip->core.height, child->core.height );
  }

  return;
}

static void
MoveChild(rw, x, y)
WWWWidget rw;
Position x, y;
{
  register Widget child = rw->www.child;
  register Widget clip = rw->www.clip;
  
  /* make sure we never move past right/bottom borders */
  if (-x + (int)clip->core.width > (int)child->core.width)
  {
    x = -(child->core.width - clip->core.width);
  }
  
  if (-y + (int)clip->core.height > (int)child->core.height)
  {
    y = -(child->core.height - clip->core.height);
  }
  
  /* make sure we never move past left/top borders */
  if (x >= 0) x = 0;
  if (y >= 0) y = 0;
  
  XtMoveWidget(child, x, y);
  
  RedrawThumbs(rw);

  return;
}

static void
ScrollUpDownProc(widget, closure, call_data)
Widget widget;
XtPointer closure;
XtPointer call_data;
{
  WWWWidget rw = (WWWWidget)XtParent(widget);
  register Widget child = rw->www.child;
  int pix = (int)call_data;
  Position x, y;
  
  if (child == NULL) return;  /* no child to scroll. */
  
  x = child->core.x - ((widget == rw->www.horiz_bar) ? pix : 0);
  y = child->core.y - ((widget == rw->www.vert_bar) ? pix : 0);
  MoveChild(rw, x, y);

  return;
}

static void
ThumbProc(widget, closure, percent)
Widget widget;
XtPointer closure;
float *percent;
{
  WWWWidget rw = (WWWWidget)XtParent(widget);
  register Widget child = rw->www.child;
  Position x, y;
  
  if (child == NULL) return;  /* no child to scroll. */
  
  if (widget == rw->www.horiz_bar)
  {
#ifdef macII                            /* bug in the macII A/UX 1.0 cc */
    x = (int)(-*percent * child->core.width);
#else /* else not macII */
    x = -(int)(*percent * child->core.width);
#endif /* macII */
  }
  else x = child->core.x;

  if (widget == rw->www.vert_bar)
  {
#ifdef macII                            /* bug in the macII A/UX 1.0 cc */
    y = (int)(-*percent * child->core.height);
#else /* else not macII */
    y = -(int)(*percent * child->core.height);
#endif /* macII */
  }
  else y = child->core.y;

  MoveChild(rw, x, y);

  return;
}

/* END SCROLL HANDLING CODE */

/*
 * WWWInitialize
 */
static void
WWWInitialize(r, n)
Widget r, n;
{
  XtGCMask valuemask;
  XGCValues values;
  Arg args[30];
  int argcnt;
  WWWWidget new = (WWWWidget)n;
  WWWWidget request = (WWWWidget)r;

  /*
   * Avoid small sizes.
   */
  if (new->core.width < 100) new->core.width = 100;
  if (new->core.height < 100) new->core.height = 100;

  if (new->www.maxDownloads < 1) new->www.maxDownloads = 1;

  /*
   * NULLs for everything.
   */
  new->www.dilist = NULL;
  new->www.iilist = NULL;
  new->www.ioqhead = NULL;
  new->www.ioqtail = NULL;
  new->www.over = NULL;
  new->www.hs = NULL;
  new->www.downloads = 0;
  new->www.rurl = NULL;
  new->www.reload = False;
  new->www.rlist = NULL;
  new->www.ips = 0;

  WWWInitLayoutVars(new);

  /*
   * Init GC
   */
  values.function = GXcopy;
  values.plane_mask = AllPlanes;
  values.foreground = new->www.foreground;
  values.background = new->www.background;
  
  valuemask = GCFunction | GCPlaneMask | GCForeground | GCBackground;
  new->www.gc = XtGetGC(n, valuemask, &values);

  /*
   * Init Image GC
   */ 
  values.function = GXcopy;
  values.plane_mask = AllPlanes;
  valuemask = GCFunction | GCPlaneMask;
  new->www.image_gc = XtGetGC(n, valuemask, &values);

  /*
   * Create scrollbars, clip, and the child widget.  The child widget is
   * the one that we draw into.
   */
  argcnt = 0;
  XtSetArg(args[argcnt], XtNorientation, XtorientVertical); argcnt++;
  XtSetArg(args[argcnt], XtNwidth, 14); argcnt++;
  new->www.vert_bar = XtCreateWidget("vert_bar",
				     scrollbarWidgetClass, n,
				     args, argcnt);
  XtAddCallback(new->www.vert_bar, XtNscrollProc,
		ScrollUpDownProc, (XtPointer)new);
  XtAddCallback(new->www.vert_bar, XtNjumpProc,
		ThumbProc, (XtPointer)new);
  
  new->www.vert_bar_width = 14;

  argcnt = 0;
  XtSetArg(args[argcnt], XtNorientation, XtorientHorizontal); argcnt++;
  XtSetArg(args[argcnt], XtNheight, 14); argcnt++;
  new->www.horiz_bar = XtCreateWidget("horiz_bar",
				      scrollbarWidgetClass, n,
				      args, argcnt);
  XtAddCallback(new->www.horiz_bar, XtNscrollProc,
		ScrollUpDownProc, (XtPointer)new);
  XtAddCallback(new->www.horiz_bar, XtNjumpProc,
		ThumbProc, (XtPointer)new);

  new->www.horiz_bar_height = 14;

  argcnt = 0;
  XtSetArg(args[argcnt], XtNwidth, 10); argcnt++;
  XtSetArg(args[argcnt], XtNheight, 10); argcnt++;
  new->www.clip = XtCreateManagedWidget("clip",
					drawingAreaWidgetClass, n,
					args, argcnt);
  
  argcnt = 0;
  XtSetArg(args[argcnt], XtNwidth, 10); argcnt++;
  XtSetArg(args[argcnt], XtNheight, 10); argcnt++;
  new->www.child = XtCreateManagedWidget("child",
					 drawingAreaWidgetClass,
					 new->www.clip,
					 args, argcnt);
  
  /*
   * Set event handlers for the child.
   */
  XtAddEventHandler(new->www.child, ExposureMask, True,
		    ExposeHandler, (XtPointer)new);
  XtAddEventHandler(new->www.child, ButtonReleaseMask, True,
		    ButtonReleaseHandler, (XtPointer)new);

  if (request->www.anchorReport)
  {
    XtAddEventHandler(new->www.child, PointerMotionMask, True,
		      PointerMotionHandler, (XtPointer)new);
  }

  MIMEReadCapFiles(new->www.mimeCapFiles);
  MIMEReadTypeFiles(new->www.mimeTypeFiles);

  return;
}

/*
 *
 * Public functions
 *
 */
void
WWWHome(w)
Widget w;
{
  WWWWidget rw = (WWWWidget)w;

  HomeDocument(rw);

  return;
}

void
WWWBack(w)
Widget w;
{
  WWWWidget rw = (WWWWidget)w;

  BackDocument(rw);

  return;
}

void
WWWReload(w)
Widget w;
{
  WWWWidget rw = (WWWWidget)w;

  ReloadDocument(rw);

  return;
}

void
WWWSearch(w, pattern, start)
Widget w;
char *pattern;
Boolean start;
{
/*  WWWWidget rw = (WWWWidget)w; */

  return;
}

void
WWWCancel(w)
Widget w;
{
  WWWWidget rw = (WWWWidget)w;

  IOStop(rw);

  return;
}
