/*
 * WWWP.h
 *
 * Copyright (c) 1995, John D. Kilburg (john@isri.unlv.edu)
 */

#ifndef __WWWP_H__
#define __WWWP_H__

#include <X11/Xcms.h>

#include "common.h"
#include "WWW.h"
#include "wwwio.h"
#include "io.h"
#include "html.h"
#include "image.h"

typedef struct
{
  XtPointer garbage;
} WWWClassPart;

typedef struct _WWWClassRec
{
  CoreClassPart	        core_class;
  CompositeClassPart	composite_class;
  ConstraintClassPart	constraint_class;
  WWWClassPart	        www_class;
} WWWClassRec;

extern WWWClassRec wwwClassRec;

/*
 * Helpful comment here
 */
typedef struct _render_env
{
  int tagid;                     /* tag ID that produced env */
  int x;                         /* x offsets */
  unsigned int width;            /* environment dimensions */
  int align;                     /* object alignment */
  int text_mode;                 /* fill, pre, title, ... */
  bool wrap;                     /* wrap lines automatically? */
  MLTag *anchorPart;             /* URL anchor */
  MLTag *namePart;               /* name anchor */
  XFontStruct *font;             /* current font */
  struct _render_env *next;      /* environment links */
} RenderEnv;

#define CurrentFont(rw)          ((rw)->www.restack->font)
#define CurrentAlign(rw)         ((rw)->www.restack->align)

typedef struct _render_part
{
  MLTag *p;                      /* */
  MLTag *anchorPart;             /* URL anchor */
  MLTag *namePart;               /* name anchor */
  int x;                         /* set by AddRenderPart, CreateRenderLine */
  int y;                         /* set by render setup proc */
  unsigned int width, height;    /* dimensions including anchor stuff */
  bool breakable;                /* allow linebreak after this part ? */
  bool renderable;               /* can the part be rendered ? */
  bool space;                    /* space */
  struct _render_part *lnext, *lprev; /* line links */
  struct _render_part *mnext;    /* next in RP master list */
  struct _render_part *anext;    /* next in anchor list */

  /* handling functions */
  void (*setup)();               /* setup function */
  void (*render)();              /* render function */
  void (*destroy)();             /* destroy function */
  void (*select)();              /* select handler */
  void *state;                   /* state for functions */
} RenderPart;

#define GetRenderState(rp)       ((rp)->state)
#define IsRenderable(rp)         ((rp)->renderable)
#define IsBreakable(rp)          ((rp)->breakable)
#define IsSpace(rp)              ((rp)->space)

typedef struct _render_line
{
  RenderPart *rphead, *rptail;   /* list of RPs */
  int x, y;                      /* position */
  int miny, maxy;                /* offset from baseline? */
  unsigned int width, height;    /* height of line */
  unsigned int maxwidth;         /* maximum width of line */
  int align;                     /* alignment */
  struct _render_line *next;     /* next line */
} RenderLine;

typedef struct _doc_info
{
  WWWWidget rw;
  Document *d;                   /* document being downloaded */
  int type;                      /* type of doc */
  float vert_shown, vert_top;    /* vertical scrollbar position */
  float horiz_shown, horiz_top;  /* horizontal scrollbar position*/
  int action;                    /* back, new, reload */
  URLParts *up;                  /* the URL of the document */
  struct _doc_info *next;        /* next document in the stack */
  int s;                         /* socket for viewer */
  int elen;                      /* keeps track of viewer data */
  bool ishtml;                   /* HTML document */
  bool inlist;                   /* in document list */
} DocInfo;

typedef struct _realminfo
{
  char *name;
  char *type;
  char *protocol;
  char *hostname;
  int port;
  char *username;
  char *password;
  int used;
  struct _realminfo *next;
} RealmInfo;

typedef struct _WWWPart
{
  /* Public */
  char        *url;
  XFontStruct *h1Font;
  XFontStruct *h2Font;
  XFontStruct *h3Font;
  XFontStruct *h4Font;
  XFontStruct *h5Font;
  XFontStruct *h6Font;
  XFontStruct *normalFont;
  XFontStruct *emFont;
  XFontStruct *bFont;
  XFontStruct *preFont;
  XFontStruct *ttFont;
  XFontStruct *iFont;
  Pixel       foreground;
  Pixel       background;
  Pixel       anchorColor;
  Dimension   margin;
  XtCallbackList anchorCallbacks;
  XtCallbackList displayCallbacks;
  XtCallbackList downloadCallbacks;
  XtCallbackList saveCallbacks;
  int         maxInlineColors;
  int         maxColors;
  int         maxDownloads;
  Dimension   lineSpace;
  Boolean     anchorReport;
  int         imageQuitInterval;
  int         ioQuantum;
  int         ioTimeOutInterval;
  char        *mimeCapFiles;
  char        *mimeTypeFiles;
  char        *path;
  Boolean     brokenHTML;

  /* Read-only resources */
  char        *rurl;                /* doc.c */
  char        *title;               /* text.c, layout.c */
  char        *author;              /* text.c, layout.c */

  /* X Private */
  Widget      vert_bar, horiz_bar;  /* scroll bars */
  Dimension   vert_bar_width;       /* vertical scrollbar width */
  Dimension   horiz_bar_height;     /* horizontal scrollbar height */
  Widget      clip;                 /* clips the child */
  Widget      child;                /* the drawing window */
  GC          gc;                   /* use this for text/line/arc drawing */
  GC          image_gc;             /* GC to use for image drawing */
  ImagePState ips;                  /* image handling info */

  /* doc.c */
  DocInfo     *dilist;              /* list of documents */
  MLState     *hs;                  /* HTML parsing stuff */

  /* io.c */
  IOInfo      *iilist;              /* list of downloads */
  IOQueue     *ioqhead, *ioqtail;   /* pending IO queue */
  int         downloads;            /* number of current downloads */

  /* WWW.c */
  RenderPart  *over;                /* pointer over this anchor */

  /* layout.c and other layout files */
  void        *formstate;           /* current form state */
  RenderPart  *aphead, *aptail;     /* RP anchor list */
  RenderPart  *rphead, *rptail;     /* RP master list */
  RenderLine  *rlhead, *rltail;     /* line list */
  RenderEnv   *restack;             /* environment stack */
  MLTag       *pqhead, *pqtail;     /* queue of unrendered tags */
  bool        delay;                /* delay layout ? */
  URLParts    *bup;                 /* base URL */
  unsigned int width, height;       /* dimensions of document */

  /* other crap */
  bool        destroy_locked;       /* deals with destroy recursion problems */
  bool        reload;               /* reloading */
  void        *memstate;            /* memstate */
  RealmInfo   *rlist;               /* list of auth info for realms
				       should be class-wide instead of here */
} WWWPart;

typedef struct _WWWRec
{
  CorePart		core;
  CompositePart     	composite;
  ConstraintPart	constraint;
  WWWPart		www;
} WWWRec;

/*
 * Private functions
 */

/*
 * form.c
 */
void HandleForm _ArgProto((WWWWidget, MLTag *, int));
void HandleInput _ArgProto((WWWWidget, MLTag *, int));
void HandleTextarea _ArgProto((WWWWidget, MLTag *, int));
void TextareaText _ArgProto((WWWWidget, MLTag *));
void OptionText _ArgProto((WWWWidget, MLTag *));
void HandleSelect _ArgProto((WWWWidget, MLTag *, int));
void HandleOption _ArgProto((WWWWidget, MLTag *, int));

/*
 * inline.c
 */
void HandleImg _ArgProto((WWWWidget, MLTag *, int));

/*
 * hr.c
 */
void HandleHR _ArgProto((WWWWidget, MLTag *, int));

/*
 * list.c
 */
void HandleLI _ArgProto((WWWWidget, MLTag *, int));
void HandleDT _ArgProto((WWWWidget, MLTag *, int));
void HandleDD _ArgProto((WWWWidget, MLTag *, int));
void HandleOL _ArgProto((WWWWidget, MLTag *, int));
void HandleList _ArgProto((WWWWidget, int, MLTag *, int));

/*
 * text.c
 */
void AuthorText _ArgProto((WWWWidget, MLTag *));
void TitleText _ArgProto((WWWWidget, MLTag *));
void FillText _ArgProto((WWWWidget, MLTag *));
void PlainText _ArgProto((WWWWidget, MLTag *, bool));
void HandlePRE _ArgProto((WWWWidget, MLTag *, int));
void HandleListing _ArgProto((WWWWidget, MLTag *, int));
void HandlePlaintext _ArgProto((WWWWidget, MLTag *, int));
void HandleTitle _ArgProto((WWWWidget, int));

/*
 * doc.c
 */
void NewDocument _ArgProto((WWWWidget, URLParts *));
void DownloadDocument _ArgProto((WWWWidget, URLParts *));
void ReloadDocument _ArgProto((WWWWidget));
void BackDocument _ArgProto((WWWWidget));
void HomeDocument _ArgProto((WWWWidget));

/*
 * layout.c
 */
void AddBlankLine _ArgProto((WWWWidget));
bool InEnv _ArgProto((WWWWidget, int));
void WWWSpacify _ArgProto((char *, int));
void WWWStartLayout _ArgProto((WWWWidget, URLParts *, bool));
void WWWEndLayout _ArgProto((WWWWidget));
void WWWLayout _ArgProto((WWWWidget, MLTag *));
URLParts *GetBaseURLParts _ArgProto((WWWWidget));
void WWWInitLayoutVars _ArgProto((WWWWidget));
int GetWidthHeight _ArgProto((MLTag *, unsigned int *, unsigned int *));

RenderPart *CreateRenderPart _ArgProto((WWWWidget, MLTag *,
					void (*)(),
					void (*)(),
					void (*)(),
					void (*)(),
					int));
int AddRenderPart _ArgProto((WWWWidget, RenderPart *,
			     unsigned int, unsigned int));
void DestroyRenderPart _ArgProto((RenderPart *));

RenderEnv *CreateRenderEnv _ArgProto((WWWWidget, int));
void DestroyRenderEnv _ArgProto((WWWWidget, int));
void DestroyRenderEnv2 _ArgProto((WWWWidget, int, bool));
void AddLineBreak _ArgProto((WWWWidget));
unsigned int RemainingLineWidth _ArgProto((WWWWidget));

/*
 * mem.c
 */
void *WWWGetMem _ArgProto((WWWWidget, int));
void WWWFreeMem _ArgProto((WWWWidget));

/*
 * event.c
 */
void ExposeHandler _ArgProto((Widget, XtPointer, XEvent *, Boolean *));
void ButtonReleaseHandler _ArgProto((Widget, XtPointer, XEvent *, Boolean *));
void PointerMotionHandler _ArgProto((Widget, XtPointer, XEvent *, Boolean *));

/*
 * auth.c
 */
bool FetchAuthInfo _ArgProto((WWWWidget, Document *, URLParts *));
void GetUserAuthInfo _ArgProto((WWWWidget, URLParts *));

/*
 * misc.c
 */
void HandleHx _ArgProto((WWWWidget, int, MLTag *, int));
void HandleP _ArgProto((WWWWidget, MLTag *, int));
void HandleBR _ArgProto((WWWWidget, MLTag *, int));
void HandleAddress _ArgProto((WWWWidget, MLTag *, int));
void HandleA _ArgProto((WWWWidget, MLTag *, int));
void HandleBase _ArgProto((WWWWidget, MLTag *, int));
void HandleBlockquote _ArgProto((WWWWidget, MLTag *, int));
void HandleFontChange _ArgProto((WWWWidget, int, int));
void HandleMisc _ArgProto((WWWWidget, int, MLTag *, int));
void HandleEOF _ArgProto((WWWWidget));

#endif /* __WWWP_H__ */
