/*
 * auth.c
 *
 * Copyright (c) 1995, John Kilburg (john@isri.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "AuthReq.h"

#include "WWWP.h"

/*
 * AuthDismissCallback
 */
static void
AuthDismissCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  URLParts *up = (URLParts *)cldata;

  DestroyURLParts(up);
  XtDestroyWidget(w);

  return;
}

/*
 * AuthOKCallback
 *
 * Called when the user enters username/password in the auth widget
 */
static void
AuthOKCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWWidget rw = (WWWWidget)XtParent(w);
  AuthReqReturnStruct *ar = (AuthReqReturnStruct *)cbdata;
  URLParts *up = (URLParts *)cldata;
  RealmInfo *r;

  up->username = alloc_string(ar->username);
  up->password = alloc_string(ar->password);

  for (r = rw->www.rlist; r; r = r->next)
  {
    if (strcasecmp(up->auth_type, r->type) == 0 &&
        strcasecmp(up->auth_realm, r->name) == 0 &&
        strcasecmp(up->protocol, r->protocol) == 0 &&
        strcasecmp(up->hostname, r->hostname) == 0 &&
        up->port == r->port)
    {
      free_mem(r->username);
      free_mem(r->password);
      r->username = alloc_string(ar->username);
      r->password = alloc_string(ar->password);
      break;
    }
  }

  if (r == NULL)
  {
    r = (RealmInfo *)alloc_mem(sizeof(RealmInfo));
    r->name = alloc_string(up->auth_realm);
    r->type = alloc_string(up->auth_type);
    r->username = alloc_string(ar->username);
    r->password = alloc_string(ar->password);
    r->protocol = alloc_string(up->protocol);
    r->hostname = alloc_string(up->hostname);
    r->port = up->port;
    r->next = rw->www.rlist;
    rw->www.rlist = r;
  }

  XtDestroyWidget(w);

  NewDocument(rw, up);

  DestroyURLParts(up);

  return;
}

/*
 * CheckRealm
 */
static Boolean
CheckRealm(rw, up)
WWWWidget rw;
URLParts *up;
{
  RealmInfo *r;

  for (r = rw->www.rlist; r != NULL; r = r->next)
  {
    if (strcasecmp(up->auth_type, r->type) == 0 &&
        strcasecmp(up->auth_realm, r->name) == 0 &&
        strcasecmp(up->protocol, r->protocol) == 0 &&
        strcasecmp(up->hostname, r->hostname) == 0 &&
        up->port == r->port)
    {
      if (up->username != NULL) free_mem(up->username);
      up->username = alloc_string(r->username);
      if (up->password != NULL) free_mem(up->password);
      up->password = alloc_string(r->password);

      return(True);
    }
  }

  return(False);
}

/*
 * FetchAuthInfo
 *
 * Returns true if the proper realm, domain, and protocol information
 * is found and the proper fields are set in URLParts.
 */
bool
FetchAuthInfo(rw, d, up)
WWWWidget rw;
Document *d;
URLParts *up;
{
  char *value;

  if ((value = MIMEGetFieldValue(d->mflist, "www-authenticate", 0)) != NULL)
  {
    char *list;
    char *t;

    list = value;
    if ((t = mystrtok(list, ' ', &list)) != NULL)
    {
      up->auth_type = alloc_string(t);
    }

    if ((t = mystrtok(list, '=', &list)) != NULL)
    {
      if (list != NULL) up->auth_realm = alloc_string(list);
    }
  }
  
  return(CheckRealm(rw, up));
}

void
GetUserAuthInfo(rw, up)
WWWWidget rw;
URLParts *up;
{
  Widget nw;
  Widget w = (Widget)rw;
  URLParts *nup;
  Window xw, cw;
  int rx, ry, wx, wy;
  unsigned int mask;

  nup = DupURLParts(up);

  XQueryPointer(XtDisplay(w), XtWindow(w),
		&xw, &cw,
		&rx, &ry,
		&wx, &wy,
		&mask);
  
  nw = XtVaCreatePopupShell("authreq",
			    authReqWidgetClass, w,
			    XtNdefaultUsername, "",
			    XtNx, rx - 2, XtNy, ry - 2,
			    NULL);
  
  XtAddCallback(nw, XtNokCallback, AuthOKCallback, (XtPointer)nup);
  XtAddCallback(nw, XtNdismissCallback, AuthDismissCallback, (XtPointer)nup);
  
  XtRealizeWidget(nw);
  XtPopup(nw, XtGrabNone);

  return;
}

