/*
 * event.c
 *
 * Copyright (c) 1995, John D. Kilburg
 *
 * See copyright.h for more information.
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#include <sys/types.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "WWWP.h"


/*
 * ExposeHandler
 */
void
ExposeHandler(w, closure, xe, dispatch)
Widget w;
XtPointer closure;
XEvent *xe;
Boolean *dispatch;
{
  WWWWidget rw = (WWWWidget)XtParent(XtParent(w));
  RenderLine *rl;
  RenderPart *rp;
  int rly2;
  Region r;
  int y1, y2;

/*  if (xe->type == NoExpose) return; */

  if (rw->www.rlhead == NULL) return;

  y1 = xe->xexpose.y;
  y2 = xe->xexpose.y + xe->xexpose.height;

  r = XCreateRegion();
  XtAddExposureToRegion(xe, r);

  for (rl = rw->www.rlhead; rl != NULL; rl = rl->next)
  {
    rly2 = rl->y + rl->height;
    if (rl->y < y1 && rly2 < y1) continue;
    else if (rl->y > y2 && rly2 > y2) return;

    for (rp = rl->rphead; rp != NULL; rp = rp->lnext)
    {
      if (IsRenderable(rp) && rp->render != NULL)
      {
	if (XRectInRegion(r, rp->x, rp->y, rp->width, rp->height))
	{
	  (*(rp->render))(rw, rp, xe);
	}
      }
    }
  }

  XDestroyRegion(r);

  return;
}

/*
 * InAnchor
 */
static Boolean
InAnchor(x, y, rp)
int x, y;
RenderPart *rp;
{
  if (x > rp->x &&
      x < rp->x + rp->width &&
      y > rp->y + rp->height - rp->height &&
      y < rp->y + rp->height)
  {
    return(True);
  }

  return(False);
}

/*
 * ButtonReleaseHandler
 */
void
ButtonReleaseHandler(w, closure, xe, dispatch)
Widget w;
XtPointer closure;
XEvent *xe;
Boolean *dispatch;
{
  WWWWidget rw = (WWWWidget)XtParent(XtParent(w));
  char *url, *aurl;
  RenderPart *rp;

  for (rp = rw->www.aphead; rp != NULL; rp = rp->anext)
  {
    if (InAnchor(xe->xbutton.x, xe->xbutton.y, rp))
    {
      if ((url = MLFindAttribute(rp->anchorPart, "href")) != NULL)
      {
	URLParts *lup;
	
	if (rp->p != NULL && MLFindAttribute(rp->p, "ismap") != NULL)
	{
	  aurl = (char *)alloc_mem(strlen(url) + BUFSIZ);
	  sprintf (aurl, "%s?%d,%d", url,
		   xe->xbutton.x - rp->x,
		   xe->xbutton.y - rp->y);
	  url = aurl;
	}
	else aurl = NULL;
	
	if ((lup = ParseURL(url)) != NULL)
	{
	  URLParts *up, *bup;
	  
	  if ((bup = GetBaseURLParts(rw)) != NULL)
	  {
	    up = MakeURLParts(lup, bup);
	    DestroyURLParts(lup);
	  }
	  else up = lup;
	  
	  if (xe->xbutton.button == 1) NewDocument(rw, up);
	  else if (xe->xbutton.button == 2) DownloadDocument(rw, up);
	  
	  DestroyURLParts(up);
	}
	
	if (aurl != NULL) free_mem(aurl);
      }
      
      return;
    }
  }

  return;
}

/*
 * PointerMotionHandler
 */
void
PointerMotionHandler(w, closure, xe, dispatch)
Widget w;
XtPointer closure;
XEvent *xe;
Boolean *dispatch;
{
  WWWWidget rw = (WWWWidget)XtParent(XtParent(w));
  char *url;
  RenderPart *rp;

  for (rp = rw->www.aphead; rp != NULL; rp = rp->anext)
  {
    if (InAnchor(xe->xbutton.x, xe->xbutton.y, rp))
    {
      if (rp == rw->www.over) return;

      if ((url = MLFindAttribute(rp->anchorPart, "href")) != NULL)
      {
	URLParts *lup;
	
	if ((lup = ParseURL(url)) != NULL)
	{
	  URLParts *up, *bup;
	  char *turl;
	  
	  if ((bup = GetBaseURLParts(rw)) != NULL)
          {
	    up = MakeURLParts(lup, bup);
	    DestroyURLParts(lup);
	  }
	  else up = lup;
	  
	  turl = MakeURL(up, 0);
	  XtCallCallbackList((Widget)rw, rw->www.anchorCallbacks,
			     (XtPointer)turl);
	  free_mem(turl);
	  DestroyURLParts(up);
	}
      }

      rw->www.over = rp;

      return;
    }
  }

  if (rw->www.over != NULL)
  {
    XtCallCallbackList((Widget)rw, rw->www.anchorCallbacks, (XtPointer)NULL);
    rw->www.over = NULL;
  }

  return;
}

