/*
 * html.c
 *
 * Copyright (c) 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "WWWP.h"

static MLAttribute *hs_parse_tag _ArgProto((MLState *, char *, int));
static void hs_create_part _ArgProto((MLState *, int,
				      char *, int,
				      char *, int,
				      MLAttribute *,
				      int));
static void hs_create_text _ArgProto((MLState *, char *, int));
static void hs_create_tag _ArgProto((MLState *, char *, int));
static char *hs_condense _ArgProto((MLState *, char *, int, int *));
static MLAttribute *hs_create_attribute _ArgProto((MLState *, char *, int));

MLTagName tname[] =
{
  { "a", TAG_A },
  { "address", TAG_ADDRESS },
  { "abbrev", TAG_ABBREV },
  { "above", TAG_ABOVE },
  { "acronym", TAG_ACRONYM },
  { "address", TAG_ADDRESS },
  { "arow", TAG_AROW },
  { "array", TAG_ARRAY },
  { "atop", TAG_ATOP },
  { "au", TAG_AU },
  { "b", TAG_B },
  { "base", TAG_BASE },
  { "below", TAG_BELOW },
  { "big", TAG_BIG },
  { "blockquote", TAG_BLOCKQUOTE },
  { "bq", TAG_BLOCKQUOTE },
  { "body", TAG_BODY },
  { "box", TAG_BOX },
  { "bq", TAG_BQ },
  { "br", TAG_BR },
  { "caption", TAG_CAPTION },
  { "cite", TAG_CITE },
  { "code", TAG_CODE },
  { "dd", TAG_DD },
  { "dfn", TAG_DFN },
  { "dir", TAG_DIR },
  { "dl", TAG_DL },
  { "dt", TAG_DT },
  { "em", TAG_EM },
  { "fig", TAG_FIG },
  { "fn", TAG_FN },
  { "form", TAG_FORM },
  { "h1", TAG_H1 },
  { "h2", TAG_H2 },
  { "h3", TAG_H3 },
  { "h4", TAG_H4 },
  { "h5", TAG_H5 },
  { "h6", TAG_H6 },
  { "head", TAG_HEAD },
  { "hr", TAG_HR },
  { "html", TAG_HTML },
  { "i", TAG_I },
  { "img", TAG_IMG },
  { "input", TAG_INPUT },
  { "isindex", TAG_ISINDEX },
  { "item", TAG_ITEM },
  { "kbd", TAG_KBD },
  { "lang", TAG_LANG },
  { "lh", TAG_LH },
  { "li", TAG_LI },
  { "link", TAG_LINK },
  { "listing", TAG_LISTING },
  { "math", TAG_MATH },
  { "menu", TAG_MENU },
  { "meta", TAG_META },
  { "nextid", TAG_NEXTID },
  { "ol", TAG_OL },
  { "option", TAG_OPTION },
  { "over", TAG_OVER },
  { "overlay", TAG_OVERLAY },
  { "p", TAG_P },
  { "person", TAG_PERSON },
  { "plaintext", TAG_PLAINTEXT },
  { "pre", TAG_PRE },
  { "q", TAG_Q },
  { "right", TAG_RIGHT },
  { "root", TAG_ROOT },
  { "s", TAG_S },
  { "samp", TAG_SAMP },
  { "select", TAG_SELECT },
  { "small", TAG_SMALL },
  { "strong", TAG_STRONG },
  { "style", TAG_STYLE },
  { "styles", TAG_STYLES },
  { "sub", TAG_SUB },
  { "sup", TAG_SUP },
  { "tab", TAG_TAB },
  { "table", TAG_TABLE },
  { "td", TAG_TD },
  { "textarea", TAG_TEXTAREA },
  { "th", TAG_TH },
  { "title", TAG_TITLE },
  { "tr", TAG_TR },
  { "tt", TAG_TT },
  { "u", TAG_U },
  { "ul", TAG_UL },
  { "var", TAG_VAR },
  { "xmp", TAG_XMP },
  { NULL, 0 },
};

/*
 * hs_create_attribute
 */
static MLAttribute *
hs_create_attribute(hs, text, len)
MLState *hs;
char *text;
int len;
{
  MLAttribute *ta;

  ta = (MLAttribute *)WWWGetMem(hs->rw, sizeof(MLAttribute));
  ta->next = NULL;

  ta->name = (char *)WWWGetMem(hs->rw, len + 1);
  strncpy(ta->name, text, len);
  ta->name[len] = '\0';

  ta->value = NULL;

  return(ta);
}

struct entity
{
  char *text;
  int len;
  unsigned char c;
} clist[] =
{
  { "amp", 3, '&' },
  { "lt", 2, '<' },
  { "gt", 2, '>' },
  { "copy", 4, 169 },
  { "AElig", 5, 198 },
  { "Aacute", 6, 193 },
  { "Acirc", 5, 194 },
  { "Agrave", 6, 192 },
  { "Aring", 5, 197 },
  { "Atilde", 6, 195 },
  { "Auml", 4, 196 },
  { "Ccedil", 6, 199 },
  { "ETH", 3, 208 },
  { "Eacute", 6, 201 },
  { "Ecirc", 5, 202 },
  { "Egrave", 6, 200 },
  { "Euml", 4, 203 },
  { "Iacute", 6, 205 },
  { "Icirc", 5, 206 },
  { "Igrave", 6, 204 },
  { "Iuml", 4, 207 },
  { "Ntilde", 6, 209 },
  { "Oacute", 6, 211 },
  { "Ocirc", 5, 212 },
  { "Ograve", 6, 210 },
  { "Oslash", 6, 216 },
  { "Otilde", 6, 213 },
  { "Ouml", 4, 214 },
  { "THORN", 5, 222 },
  { "Uacute", 6, 218 },
  { "Ucirc", 5, 219 },
  { "Ugrave", 6, 217 },
  { "Uuml", 4, 220 },
  { "Yacute", 6, 221 },
  { "aacute", 6, 225 },
  { "acirc", 5, 226 },
  { "aelig", 5, 230 },
  { "agrave", 6, 224 },
  { "aring", 5, 229 },
  { "atilde", 6, 227 },
  { "auml", 4, 228 },
  { "ccedil", 6, 231 },
  { "eacute", 6, 233 },
  { "ecirc", 5, 234 },
  { "egrave", 6, 232 },
  { "eth", 3, 240 },
  { "euml", 4, 235 },
  { "iacute", 6, 237 },
  { "icirc", 5, 238 },
  { "igrave", 6, 236 },
  { "iuml", 4, 239 },
  { "ntilde", 6, 241 },
  { "oacute", 6, 243 },
  { "ocirc", 5, 244 },
  { "ograve", 6, 242 },
  { "oslash", 6, 248 },
  { "otilde", 6, 245 },
  { "ouml", 4, 246 },
  { "szlig", 5, 223 },
  { "thorn", 5, 254 },
  { "uacute", 6, 250 },
  { "ucirc", 5, 251 },
  { "ugrave", 6, 249 },
  { "uuml", 4, 252 },
  { "yacute", 6, 253 },
  { "yuml", 4, 255 },
  { "reg", 3, 174 },
  { "comma", 5, 44 },
  { "colon", 5, 58 },
  { "quot", 4, '\"' },
  { "nbsp", 4, ' ' },
  { NULL, 0, '\0' },
};

/*
 * hs_condense
 */
static char *
hs_condense(hs, text, len, newlen)
MLState *hs;
char *text;
int len;
int *newlen;
{
  char *cp, *lastcp;
  int i;
  char *b, *bcp;
  char *x;

  /*
   * allocate a buffer for the new text.  the new text will be smaller than
   * the incoming text
   */
  b = (char *)WWWGetMem(hs->rw, len + 1);
  bcp = b;

  for (cp = text, lastcp = text + len; cp < lastcp; cp++)
  {
    /*
     * Found entity and there is at least one more character?
     */
    if (*cp == '&')
    {
      for (x = cp; x < lastcp && *x != ';' && !isspace(*x); x++)
	  ;
      if (x < lastcp)                              /* found end of entity ? */
      {
	if (*x == '\r') x++;                       /* this may die */
	if (*(cp + 1) == '#')                      /* number given ? */
	{
	  *bcp = (char)atoi(cp + 2);               /* get the character */
	  bcp++;
	  cp = x;
	}
	else
	{
	  char *crap;

	  /* Search in the table of entity names */
	  for (i = 0, crap = cp + 1; clist[i].text != NULL; i++)
	  {
	    if (strncmp(clist[i].text, crap, clist[i].len) == 0)
	    {
	      *bcp = clist[i].c;
	      bcp++;
	      cp = x;
	      break;
	    }
	  }
	}
      }
      if (cp < x)
      {
	*bcp = *cp;                 
	bcp++;
      }
    }
    else
    {
      *bcp = *cp;                                 /* not part of entity */
      bcp++;
    }
  }

  *newlen = bcp - b;
  *bcp = '\0';

  return(b);
}

/*
 * hs_parse_tag
 */
MLAttribute *
hs_parse_tag(hs, text, len)
MLState *hs;
char *text;
int len;
{
  char *cp, *start, *lastcp;
  MLAttribute *tahead, *tatail, *ta;
  int newlen;

  tahead = NULL;
  tatail = NULL;
  lastcp = text + len;
  for (cp = text; cp < lastcp; )
  {
    /* Eat leading spaces */
    for (; cp < lastcp && isspace(*cp); cp++)
	;
    if (cp == lastcp) break;

    /* Look for value */
    for (start = cp; cp < lastcp && !isspace(*cp) && *cp != '='; cp++)
	;

    ta = hs_create_attribute(hs, start, cp - start);

    if (isspace(*cp))
    {
      for (; cp < lastcp && isspace(*cp); cp++)
	  ;
    }

    /* Found value */
    if (cp < lastcp && *cp == '=')
    {
      cp++; /* past '=' */

      /* Eat leading spaces */
      for (; cp < lastcp && isspace(*cp); cp++)
	  ;
      if (cp < lastcp)
      {
	/* Quoted value */
	if (*cp == '"')
	{
	  cp++; /* past '"' */

	  for (start = cp; cp < lastcp && *cp != '"'; cp++)
	      ;
	}
	else if (*cp == '\'') /* single quoted value */
	{
	  cp++; /* past ' */

	  for (start = cp; cp < lastcp && *cp != '\''; cp++)
	      ;
	}
	else /* Unquoted value */
	{
	  for (start = cp; cp < lastcp && !isspace(*cp); cp++)
	      ;
	}

	ta->value = hs_condense(hs, start, cp - start, &newlen);
      }
    }

    if (tatail == NULL) tahead = ta;
    else tatail->next = ta;
    tatail = ta;
  }

  return(tahead);
}


/*
 * hs_create_part
 */
static void
hs_create_part(hs, tagid, otext, olen, text, len, ta, end)
MLState *hs;
int tagid;
char *otext;
int olen;
char *text;
int len;
MLAttribute *ta;
int end;
{
  MLTag *p;

  p = (MLTag *)WWWGetMem(hs->rw, sizeof(MLTag));
  p->id = tagid;
  p->end = end;
  p->otext = otext;
  p->olen = olen;
  p->text = text;
  p->len = len;
  p->ta = ta;
  p->next = NULL;
  p->qnext = NULL;

  if (hs->ptail != NULL) hs->ptail->next = p;
  else hs->phead = p;
  hs->ptail = p;

  (*(hs->func))(hs->rw, p);

  return;
}

/*
 * hs_create_text
 *
 * This thing needs to be rewritten to make separate parts for each
 * entity so that memory usage will not get out of hand.
 */
static void
hs_create_text(hs, text, len)
MLState *hs;
char *text;
int len;
{
  char *cp, *lastcp = text + len;
  int newlen;

  for (cp = text; cp < lastcp && *cp != '&'; cp++)
      ;

  if (cp == lastcp) hs_create_part(hs, TAG_TEXT, text, len,
				   text, len, NULL, 0);
  else
  {
    cp = hs_condense(hs, text, len, &newlen);
    hs_create_part(hs, TAG_TEXT, text, len, cp, newlen, NULL, 0);
  }

  return;
}

/*
 * hs_create_tag
 */
static void
hs_create_tag(hs, text, len)
MLState *hs;
char *text;
int len;
{
  char *name;
  int end;
  int i;
  MLAttribute *ta;

  if ((ta = hs_parse_tag(hs, text + 1, len - 2)) == NULL) return;

  if (*(ta->name) == '/')
  {
    name = ta->name + 1;
    end = 1;
  }
  else
  {
    name = ta->name;
    end = 0;
  }

  for (i = 0; tname[i].name != NULL; i++)
  {
    if (strcasecmp(tname[i].name, name) == 0)
    {
      hs_create_part(hs, tname[i].id, text, len, NULL, 0, ta, end);
      return;
    }
  }

  hs_create_part(hs, TAG_UNKNOWN, text, len, NULL, 0, ta, end);

  return;
}

/*
 * MLInit
 */
MLState *
MLInit(rw, func)
WWWWidget rw;
void (*func)();
{
  MLState *hs;

  hs = (MLState *)WWWGetMem(rw, sizeof(MLState));
  hs->rw = rw;
  hs->data = NULL;
  hs->len = 0;
  hs->cp = NULL;
  hs->phead = NULL;
  hs->ptail = NULL;
  hs->tags = NULL;
  hs->texts = NULL;
  hs->func = func;
  hs->inquotes = 0;
  hs->incomment = 0;

  return(hs);
}

/*
 * MLDestroy
 */
void
MLDestroy(hs)
MLState *hs;
{
  return;
}

/*
 * MLAddData
 */
void
MLAddData(hs, data, len)
MLState *hs;
char *data;
int len;
{
  char *lastcp;

  hs->data = data;
  hs->len = len;

  if (hs->cp == NULL)
  {
    hs->cp = data;
    if (*data != '<') hs->texts = data;
  }

  for (lastcp = data + len; hs->cp < lastcp; hs->cp++)
  {
    if (*hs->cp == '<' && hs->inquotes == 0 && hs->incomment == 0)
    {
      if ((hs->cp + 1) == lastcp) return;
      else if (*(hs->cp + 1) == '!')
      {
	if ((hs->cp + 2) == lastcp) return;
	else if (*(hs->cp + 2) == '-')
	{
	  if ((hs->cp + 3) == lastcp) return;
	  else if (*(hs->cp + 3) == '-') hs->incomment = 1;
	}
      }

      if (hs->texts != NULL)
      {
	hs_create_text(hs, hs->texts, hs->cp - hs->texts);
	hs->texts = NULL;
      }
      hs->tags = hs->cp;
    }
    else if (*hs->cp == '>' && hs->inquotes == 0 && hs->incomment == 0)
    {
      if (hs->tags != NULL)
      {
	hs_create_tag(hs, hs->tags, hs->cp - hs->tags + 1);
	hs->tags = NULL;
      }
      hs->texts = hs->cp + 1;
    }
    else if (hs->incomment != 0 && hs->inquotes == 0 && *hs->cp == '-')
    {
      if ((hs->cp + 1) == lastcp) return;
      else if (*(hs->cp + 1) == '-')
      {
	if ((hs->cp + 2) == lastcp) return;
	else if (*(hs->cp + 2) == '>')
        {
	  hs->incomment = 0;
	  hs->cp += 2;

	  if (hs->tags != NULL)
	  {
	    hs_create_tag(hs, hs->tags, hs->cp - hs->tags + 1);
	    hs->tags = NULL;
	  }
	  hs->texts = hs->cp + 1;
	}
      }
    }
    else if (*hs->cp == '"' && hs->tags != NULL)
    {
      if (!hs->rw->www.brokenHTML) hs->inquotes = 1 - hs->inquotes;
    }
  }

  return;
}

/*
 * MLEndData
 */
void
MLEndData(hs)
MLState *hs;
{
  char *lastcp = hs->data + hs->len;
    
  if (hs->texts != NULL)
  {
    hs_create_text(hs, hs->texts, lastcp - hs->texts);
  }
  else if (hs->tags != NULL)
  {
    hs_create_tag(hs, hs->tags, lastcp - hs->tags);
  }
  
  hs_create_part(hs, TAG_EOF, NULL, 0, NULL, 0, NULL, 0);
  
  return;
}

/*
 * MLFindAttribute
 */
char *
MLFindAttribute(p, name)
MLTag *p;
char *name;
{
  MLAttribute *ta;

  for (ta = p->ta; ta != NULL; ta = ta->next)
  {
    if (strcasecmp(ta->name, name) == 0)
    {
      return(ta->value == NULL ? "":ta->value);
    }
  }

  return(NULL);
}

/*
 * MLGetAlignID
 */
int
MLGetAlignID(s)
char *s;
{
  int align_id;

  if (strcasecmp(s, "left") == 0) align_id = ALIGN_LEFT;
  else if (strcasecmp(s, "center") == 0) align_id = ALIGN_CENTER;
  else if (strcasecmp(s, "right") == 0) align_id = ALIGN_RIGHT;
  else if (strcasecmp(s, "top") == 0) align_id = ALIGN_TOP;
  else if (strcasecmp(s, "bottom") == 0) align_id = ALIGN_BOTTOM;
  else align_id = ALIGN_JUSTIFY;

  return(align_id);
}
