/*
 * pipe.c
 *
 * Copyright (C) 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>

#include <fcntl.h>

#if ((defined(SYSV) || defined(SVR4)) && defined(sun))
#include <sys/file.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "wwwio.h"

/*
 * PipeCommand
 *
 * fork and exec to get a program running and supply it with
 * a stdin, stdout, stderr that so we can talk to it.
 */
int
PipeCommand(command, fd, usesystem)
char *command;
int *fd;
int usesystem;
{
  int pout[2];
  int pin[2];
  int pid;

  if (pipe(pout) == -1) return(-1);

  if (pipe(pin) == -1)
  {
    close(pout[0]);
    close(pout[1]);
    return(-1);
  }

  pid = fork();
  if (pid == -1)
  {
    return(-1);
  }
  else if (pid == 0)
  {
#ifdef CHILD_STDERR
    int perr;
    char *filename;

    filename = FixFilename(CHILD_STDERR);
    if (filename == NULL) filename = "/dev/null";

    perr = open(filename, O_WRONLY | O_CREAT, 0);
    if (perr >= 0)
    {
      dup2(perr, 2);
      close(perr);
    }
#endif

    if (pout[1] != 1)
    {
      dup2(pout[1], 1);
      close(pout[1]);
    }

    if (pin[0] != 0)
    {
      dup2(pin[0], 0);
      close(pin[0]);
    }

    if (usesystem)
    {
      system(command);
      exit(0);
    }
    else
    {
      execl(command, command, (char *)0);
      exit(1);
    }
  }
  else
  {
    close(pout[1]);
    close(pin[0]);
  }

  fd[0] = pin[1];
  fd[1] = pout[0];

  fcntl(fd[0], F_SETFL, FNDELAY);
  fcntl(fd[1], F_SETFL, FNDELAY);

  return(0);
}
