/*
 * www.h
 *
 * Copyright (C) 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#ifndef __WWWIO_H__
#define __WWWIO_H__ 1

#include <time.h> /* for time_t */

#include "url.h"

typedef struct _mimefield
{
  char *name;
  char *value;
  int used;
  struct _mimefield *next;
} MIMEField;

typedef struct _mimecap
{
  char *ctype;
  char *command;
  struct _mimecap *next;  
} MIMECap;

#define DS_DOWNLOADING 0
#define DS_NEEDS_AUTH 1
#define DS_ERROR 2
#define DS_NOTHING 3
#define DS_TIMEOUT 4
#define DS_ABORT 5
#define DS_FINISHED 6

typedef struct _document
{
  void *data;           /* address of the document data */
  int len;              /* the len of the data */
  MIMEField *mflist;    /* list of MIME fields for document */
} Document;

typedef void *IOContext;

/*
 * convert.c
 */
void ReadConvertFiles _ArgProto((char *));
void ConvertDocument _ArgProto((IOContext, char *, char *, Document *,
				int (*)(), void *));

/*
 * document.c
 */
void ReadProtocolFiles _ArgProto((char *));
void LoadDocument _ArgProto((IOContext, URLParts *, int (*)(), void *));

void DestroyDocument _ArgProto((Document *));
Document *CreateDocument _ArgProto((void));
Document *BuildDocument _ArgProto((void *, int, char *, int));

/*
 * mime.c
 */
char *MIMEExt2Content _ArgProto((char *));
void MIMEReadTypeFiles _ArgProto((char *));
void MIMEReadCapFiles _ArgProto((char *));
int MIMEStartViewer _ArgProto((MIMECap *, char *));
MIMECap *MIMEFindViewer _ArgProto((char *));
void MIMEParseHeaderLine _ArgProto((Document *, char *));
void MIMEAddHeaderField _ArgProto((Document *, char *, char *));
void MIMEDestroyField _ArgProto((MIMEField *));
char *MIMEGetFieldValue _ArgProto((MIMEField *, char *, int));
time_t ParseExpiresDate _ArgProto((char *));

/*
 * pipe.c
 */
int PipeCommand _ArgProto((char *, int *, int));

/*
 * Protocol functions: http.c, gopher.c, local.c, ftp.c
 */
void http _ArgProto((IOContext, URLParts *, int (*)(), void *));

void http_proxy _ArgProto((IOContext, URLParts *, URLParts *,
			   int (*)(), void *));
void http_local _ArgProto((IOContext, URLParts *, char *,
			   int (*)(), void *));

void gopher _ArgProto((IOContext, URLParts *, int (*)(), void *));
void gopherp _ArgProto((IOContext, URLParts *, int (*)(), void *));
void file _ArgProto((IOContext, URLParts *, int (*)(), void *));
void ftp _ArgProto((IOContext, URLParts *, int (*)(), void *));

#endif /* __WWWIO_H__ */
